//
// async_tcp_echo_server_customalloc.cpp
// ~~~~~~~~~~~~~~~~~~~~~~~~~
//
// Copyright (c) 2003-2017 Christopher M. Kohlhoff (chris at kohlhoff dot com)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//

#include <cstdlib>
#include <iostream>
#include "fixed_pool_allocators.h"
#include <boost/aligned_storage.hpp>
#include <boost/array.hpp>
#include <boost/bind.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <boost/asio.hpp>

using boost::asio::ip::tcp;

class session
	: public boost::enable_shared_from_this<session>
{
public:
	session(boost::asio::io_service& io_service)
		: socket_(io_service),
			readHandlerPool_(1), readAllocator_(readHandlerPool_),
			writeHandlerPool_(1), writeAllocator_(writeHandlerPool_)
	{
	}

	tcp::socket& socket()
	{
		return socket_;
	}

	void start()
	{
		socket_.async_read_some(boost::asio::buffer(data_),
				custom_alloc_async_read_some_handler<async_read_some_handler>(readAllocator_,
					boost::bind(&session::handle_read,
						shared_from_this(),
						boost::asio::placeholders::error,
						boost::asio::placeholders::bytes_transferred)));
	}

private:
	void handle_read(const boost::system::error_code& error,
			size_t bytes_transferred)
	{
		if (!error)
		{
			boost::asio::async_write(socket_,
					boost::asio::buffer(data_, bytes_transferred),
					custom_alloc_async_write_handler<async_write_handler>(writeAllocator_,
						boost::bind(&session::handle_write,
							shared_from_this(),
							boost::asio::placeholders::error)));
		}
	}

	void handle_write(const boost::system::error_code& error)
	{
		if (!error)
		{
			socket_.async_read_some(boost::asio::buffer(data_),
					custom_alloc_async_read_some_handler<async_read_some_handler>(readAllocator_,
						boost::bind(&session::handle_read,
							shared_from_this(),
							boost::asio::placeholders::error,
							boost::asio::placeholders::bytes_transferred)));
		}
	}

private:
	// The socket used to communicate with the client.
	tcp::socket socket_;

	// Buffer used to store data received from the client.
	boost::array<char, 1024> data_;

	// The allocators to use for handler-based custom memory allocation.
	typedef boost::_bi::bind_t<void,
		boost::_mfi::mf2<void,session,boost::system::error_code const &,unsigned int>,
		boost::_bi::list3<boost::_bi::value<boost::shared_ptr<session> >,
		boost::arg<1>,boost::arg<2> > > async_read_some_handler;
	async_read_some_handler_allocator<async_read_some_handler>::pool_type readHandlerPool_;
	async_read_some_handler_allocator<async_read_some_handler> readAllocator_;
	typedef boost::_bi::bind_t<void,
		boost::_mfi::mf1<void,session,boost::system::error_code const &>,
		boost::_bi::list2<boost::_bi::value<boost::shared_ptr<session> >,
		boost::arg<1> > > async_write_handler;
	async_write_handler_allocator<async_write_handler>::pool_type writeHandlerPool_;
	async_write_handler_allocator<async_write_handler> writeAllocator_;
};

typedef boost::shared_ptr<session> session_ptr;

class server
{
public:
	server(boost::asio::io_service& io_service, short port)
		: io_service_(io_service),
			acceptor_(io_service, tcp::endpoint(tcp::v4(), port))
	{
		start_accept();
	}

private:
	void start_accept()
	{
		session_ptr new_session(new session(io_service_));
		acceptor_.async_accept(new_session->socket(),
				boost::bind(&server::handle_accept, this, new_session,
					boost::asio::placeholders::error));
	}

	void handle_accept(session_ptr new_session,
			const boost::system::error_code& error)
	{
		if (!error)
		{
			new_session->start();
		}

		start_accept();
	}

private:
	boost::asio::io_service& io_service_;
	tcp::acceptor acceptor_;
};

int main(int argc, char* argv[])
{
	try
	{
		if (argc != 2)
		{
			std::cerr << "Usage: async_tcp_echo_server_customalloc <port>\n";
			return 1;
		}

		boost::asio::io_service io_service;

		using namespace std; // For atoi.
		server s(io_service, atoi(argv[1]));

		io_service.run();
	}
	catch (std::exception& e)
	{
		std::cerr << "Exception: " << e.what() << "\n";
	}

	return 0;
}
