#ifndef FIXED_POOL_ALLOCATORS_H
#define FIXED_POOL_ALLOCATORS_H

#include <boost/assert.hpp>
#include <boost/noncopyable.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>
#include <boost/asio/socket_base.hpp>
#include <boost/asio/io_service.hpp>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/write.hpp>
#include <boost/asio/basic_stream_socket.hpp>
#include <boost/asio/detail/reactive_socket_recv_op.hpp>
#include <boost/asio/detail/reactive_socket_send_op.hpp>

template<typename T>
class fixed_object_pool
	: private boost::noncopyable
{
private:
	char *rawStorage;
	T *storage;
	size_t storageSize;
	T *nextFree;
public:
	fixed_object_pool(size_t S)
		: storage(nullptr), nextFree(nullptr)
	{
		// Allocate memory.
		rawStorage = new char[sizeof(T) * S];
		storage = reinterpret_cast<T*>( rawStorage );
		storageSize = S;

		// All objects are free, initially.
		for (size_t i = 0; i < S; ++i)
		{
			T **pFreeObject = reinterpret_cast<T**>(&storage[i]);
			*pFreeObject = nextFree;
			nextFree = reinterpret_cast<T*>(&storage[i]);
		}
	}

	~fixed_object_pool()
	{
		// TODO: Verify that all allocated objects have been freed.
		delete[] rawStorage;
	}

	T *allocate()
	{
		T *freeObject = nextFree;
		if (nextFree != nullptr)
		{
			T **pFreeObject = reinterpret_cast<T**>(nextFree);
			nextFree = *pFreeObject;
		}
		return freeObject;
	}

	void deallocate(T *obj)
	{
		BOOST_ASSERT(obj >= storage && obj < storage + storageSize);
		T **pFreeObject = reinterpret_cast<T**>(obj);
		*pFreeObject = nextFree;
		nextFree = obj;
	}
};

template<typename T>
class shared_object_allocator : public boost::detail::sp_ms_deleter<T>
{
public:
	typedef boost::detail::sp_counted_impl_pda<T *, boost::detail::sp_ms_deleter<T>, shared_object_allocator<T> > shared_type;
	typedef fixed_object_pool<shared_type> pool_type;
private:
	pool_type &m_rPool;
	void operator=(shared_object_allocator const &other);
public:
	explicit shared_object_allocator(pool_type &a_rPool) : m_rPool(a_rPool) { }
	shared_object_allocator(shared_object_allocator const &other) : m_rPool(other.m_rPool) { }

	template <typename U>
	struct rebind
	{
		typedef shared_object_allocator/*<U>*/ other;
	};

	shared_type *allocate(int iCount)
	{
		BOOST_ASSERT(iCount == 1);
		return m_rPool.allocate();
	}
	void deallocate(shared_type *obj, size_t iCount)
	{
		BOOST_ASSERT(iCount == 1);
		m_rPool.deallocate(obj);
	}
};

template<typename T>
class handler_allocator
{
public:
	typedef T handler_type;
	typedef fixed_object_pool<handler_type> pool_type;
private:
	pool_type &m_rPool;
	void operator=(handler_allocator const &other);
public:
	explicit handler_allocator(pool_type &a_rPool) : m_rPool(a_rPool) { }
	handler_allocator(handler_allocator const &other) : m_rPool(other.m_rPool) { }

	template <typename U>
	struct rebind
	{
		typedef handler_allocator/*<U>*/ other;
	};

	handler_type *allocate(int iCount)
	{
		BOOST_ASSERT(iCount == 1);
		return m_rPool.allocate();
	}
	void deallocate(handler_type *obj, size_t iCount)
	{
		BOOST_ASSERT(iCount == 1);
		m_rPool.deallocate(obj);
	}
};

// Class to manage the memory to be used for handler-based custom allocation.
// It contains a single block of memory which may be returned for allocation
// requests.
template<typename T>
class async_read_some_handler_allocator
{
public:
	typedef boost::asio::detail::reactive_socket_recv_op<boost::asio::mutable_buffers_1, T> handler_type;
	typedef fixed_object_pool<handler_type> pool_type;
private:
	pool_type &m_rPool;
	void operator=(async_read_some_handler_allocator const &other);
public:
	explicit async_read_some_handler_allocator(pool_type &a_rPool) : m_rPool(a_rPool) { }
	async_read_some_handler_allocator(async_read_some_handler_allocator const &other) : m_rPool(other.m_rPool) { }

	template <typename U>
	struct rebind
	{
		typedef async_read_some_handler_allocator/*<U>*/ other;
	};

	handler_type *allocate(int iCount)
	{
		BOOST_ASSERT(iCount == 1);
		return m_rPool.allocate();
	}
	void deallocate(handler_type *obj, size_t iCount)
	{
		BOOST_ASSERT(iCount == 1);
		m_rPool.deallocate(obj);
	}
};

// Class to manage the memory to be used for handler-based custom allocation.
// It contains a single block of memory which may be returned for allocation
// requests.
template<typename T>
class async_write_handler_allocator
{
public:
	typedef boost::asio::detail::reactive_socket_send_op<boost::asio::mutable_buffers_1,
		boost::asio::detail::write_op<boost::asio::basic_stream_socket<boost::asio::ip::tcp,
		boost::asio::stream_socket_service<boost::asio::ip::tcp> >,boost::asio::mutable_buffers_1,
		boost::asio::detail::transfer_all_t, T> > handler_type;
	typedef fixed_object_pool<handler_type> pool_type;
private:
	pool_type &m_rPool;
	void operator=(async_write_handler_allocator const &other);
public:
	explicit async_write_handler_allocator(pool_type &a_rPool) : m_rPool(a_rPool) { }
	async_write_handler_allocator(async_write_handler_allocator const &other) : m_rPool(other.m_rPool) { }

	template <typename U>
	struct rebind
	{
		typedef async_write_handler_allocator/*<U>*/ other;
	};

	handler_type *allocate(int iCount)
	{
		BOOST_ASSERT(iCount == 1);
		return m_rPool.allocate();
	}
	void deallocate(handler_type *obj, size_t iCount)
	{
		BOOST_ASSERT(iCount == 1);
		m_rPool.deallocate(obj);
	}
};

// Wrapper class template for handler objects to allow handler memory
// allocation to be customised. Calls to operator() are forwarded to the
// encapsulated handler.
template <typename Handler>
class custom_alloc_handler
{
public:
	custom_alloc_handler(handler_allocator<Handler>& a, Handler h)
		: allocator_(a),
			handler_(h)
	{
	}

	template <typename Arg1>
	void operator()(Arg1 arg1)
	{
		handler_(arg1);
	}

	template <typename Arg1, typename Arg2>
	void operator()(Arg1 arg1, Arg2 arg2)
	{
		handler_(arg1, arg2);
	}

	friend void* asio_handler_allocate(std::size_t size,
			custom_alloc_handler<Handler>* this_handler)
	{
		BOOST_ASSERT (size == sizeof(Handler));
		return this_handler->allocator_.allocate(1);
	}

	friend void asio_handler_deallocate(void* pointer, std::size_t size,
			custom_alloc_handler<Handler>* this_handler)
	{
		BOOST_ASSERT (size == sizeof(Handler));
		this_handler->allocator_.deallocate((Handler*)pointer, 1);
	}

private:
	handler_allocator<Handler>& allocator_;
	Handler handler_;
};

// Wrapper class template for handler objects to allow handler memory
// allocation to be customised. Calls to operator() are forwarded to the
// encapsulated handler.
template <typename Handler>
class custom_alloc_async_read_some_handler
{
public:
	typedef boost::asio::detail::reactive_socket_recv_op<boost::asio::mutable_buffers_1, Handler> handler_type;

	custom_alloc_async_read_some_handler(async_read_some_handler_allocator<Handler>& a, Handler h)
		: allocator_(a),
			handler_(h)
	{
	}

	template <typename Arg1>
	void operator()(Arg1 arg1)
	{
		handler_(arg1);
	}

	template <typename Arg1, typename Arg2>
	void operator()(Arg1 arg1, Arg2 arg2)
	{
		handler_(arg1, arg2);
	}

	friend void* asio_handler_allocate(std::size_t size,
			custom_alloc_async_read_some_handler<Handler>* this_handler)
	{
		BOOST_ASSERT (size == sizeof(handler_type));
		return this_handler->allocator_.allocate(1);
	}

	friend void asio_handler_deallocate(void* pointer, std::size_t size,
			custom_alloc_async_read_some_handler<Handler>* this_handler)
	{
		BOOST_ASSERT (size == sizeof(handler_type));
		this_handler->allocator_.deallocate((handler_type*)pointer, 1);
	}

private:
	async_read_some_handler_allocator<Handler>& allocator_;
	Handler handler_;
};

// Wrapper class template for handler objects to allow handler memory
// allocation to be customised. Calls to operator() are forwarded to the
// encapsulated handler.
template <typename Handler>
class custom_alloc_async_write_handler
{
public:
	typedef boost::asio::detail::reactive_socket_send_op<boost::asio::mutable_buffers_1,
		boost::asio::detail::write_op<boost::asio::basic_stream_socket<boost::asio::ip::tcp,
		boost::asio::stream_socket_service<boost::asio::ip::tcp> >,boost::asio::mutable_buffers_1,
		boost::asio::detail::transfer_all_t, Handler> > handler_type;

	custom_alloc_async_write_handler(async_write_handler_allocator<Handler>& a, Handler h)
		: allocator_(a),
			handler_(h)
	{
	}

	template <typename Arg1>
	void operator()(Arg1 arg1)
	{
		handler_(arg1);
	}

	template <typename Arg1, typename Arg2>
	void operator()(Arg1 arg1, Arg2 arg2)
	{
		handler_(arg1, arg2);
	}

	friend void* asio_handler_allocate(std::size_t size,
			custom_alloc_async_write_handler<Handler>* this_handler)
	{
		BOOST_ASSERT (size == sizeof(handler_type));
		return this_handler->allocator_.allocate(1);
	}

	friend void asio_handler_deallocate(void* pointer, std::size_t size,
			custom_alloc_async_write_handler<Handler>* this_handler)
	{
		BOOST_ASSERT (size == sizeof(handler_type));
		this_handler->allocator_.deallocate((handler_type*)pointer, 1);
	}

private:
	async_write_handler_allocator<Handler>& allocator_;
	Handler handler_;
};

#endif // FIXED_POOL_ALLOCATORS_H
